# -*- coding: utf-8 -*-
###############################################################################~!!~~!!!
#
# Copyright (c) 2010-2012, OPENTIA Group (<http://opentia.com>)
# The word "OPENTIA" is an European Community Trademark property of the Opentia Group
#
# @author: Opentia "Happy Hacking" Team
# @e-mail: consultoria@opentia·es
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation, either version 3 of the
#    License, or (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
###############################################################################~!!~~!!!


from osv import fields, osv
import decimal_precision as dp
from operator import itemgetter
from itertools import groupby

class stock_location_extension(osv.osv):
    _inherit = "stock.location"

    def _product_value(self, cr, uid, ids, field_names, arg, context=None):
        """Computes stock value (real and virtual) for a product, as well as stock qty (real and virtual).
        @param field_names: Name of field
        @return: Dictionary of values
        """
        prod_id = context and context.get('product_id', False)

        product_product_obj = self.pool.get('product.product')
        #TODO: ORM?
        cr.execute('select distinct product_id, location_id from stock_move where location_id in %s', (tuple(ids), ))
        dict1 = cr.dictfetchall()
        cr.execute('select distinct product_id, location_dest_id as location_id from stock_move where location_dest_id in %s', (tuple(ids), ))
        dict2 = cr.dictfetchall()
        res_products_by_location = sorted(dict1+dict2, key=itemgetter('location_id'))
        products_by_location = dict((k, [v['product_id'] for v in itr]) for k, itr in groupby(res_products_by_location, itemgetter('location_id')))

        result = dict([(i, {}.fromkeys(field_names, 0.0)) for i in ids])
        result.update(dict([(i, {}.fromkeys(field_names, 0.0)) for i in list(set([aaa['location_id'] for aaa in res_products_by_location]))]))

        currency_id = self.pool.get('res.users').browse(cr, uid, uid).company_id.currency_id.id
        currency_obj = self.pool.get('res.currency')
        currency = currency_obj.browse(cr, uid, currency_id, context=context)
        for loc_id, product_ids in products_by_location.items():
            if prod_id:
                product_ids = [prod_id]
            c = (context or {}).copy()
            c['location'] = loc_id
            for prod in product_product_obj.browse(cr, uid, product_ids, context=c):
                for f in field_names:
                    if f == 'stock_real':
                        if loc_id not in result:
                            result[loc_id] = {}
                        result[loc_id][f] += prod.qty_available
                    elif f == 'stock_virtual':
                        result[loc_id][f] += prod.virtual_available
                    elif f == 'stock_real_value':
                        amount = prod.qty_available * prod.standard_price
                        amount = currency_obj.round(cr, uid, currency, amount)
                        result[loc_id][f] += amount
                    elif f == 'stock_virtual_value':
                        amount = prod.virtual_available * prod.standard_price
                        amount = currency_obj.round(cr, uid, currency, amount)
                        result[loc_id][f] += amount
                    # Adds
                    elif f == 'stock_real_available_qty':
                        if loc_id not in result:
                            result[loc_id] = {}
                        result[loc_id][f] += prod.real_available_qty
                    elif f == 'stock_virtual_available_qty':
                        result[loc_id][f] += prod.virtual_available_qty
        return result

    _columns = {
        'in_move_ids': fields.one2many('stock.move', 'location_dest_id', 'In Moves', readonly=True),
        'out_move_ids': fields.one2many('stock.move', 'location_id', 'Out Moves', readonly=True),
        'stock_real_available_qty': fields.function(_product_value, method=True, type='float', string='Real Quantity Available', help="Real stock without locked units.", multi='stock'),
        'stock_virtual_available_qty': fields.function(_product_value, method=True,  type='float', string='Virtual Quantity Available', help="Virtual stock without locked units.", multi='stock'),
    }
stock_location_extension()

class available_product_stock(osv.osv):
    _inherit = "product.product"

    def get_product_available(self, cr, uid, ids, context=None):
        """ Finds whether product is available or not in particular warehouse.
        @return: Dictionary of values
        """
        if context is None: context = {}
        states = context.get('states',[])
        what = context.get('what',())
        if not ids: ids = self.search(cr, uid, [])
        res = {}.fromkeys(ids, 0.0)
        if not ids: return res

        if context.get('shop', False):
            shop_ids = self.pool.get('sale.shop').search(cr,uid,[('id','=',int(context['shop']))], context=context)
            shop = self.pool.get('sale.shop').browse(cr,uid, shop_ids, context=context)[0]
            if shop.id: context['warehouse'] = shop.id

        if context.get('warehouse', False):
            warehouse_ids = self.pool.get('stock.warehouse').search(cr,uid,[('id','=',int(context['warehouse']))], context=context)
            warehouse = self.pool.get('stock.warehouse').browse(cr,uid, warehouse_ids, context=context)[0]
            if warehouse.id: context['location'] = warehouse.id

        if context.get('location', False):
            if type(context['location']) == type(1): location_ids = [context['location']]
            elif type(context['location']) in (type(''), type(u'')):
                location_ids = self.pool.get('stock.location').search(cr, uid, [('name','ilike',context['location'])], context=context)
            else: location_ids = context['location']
        else:
            location_ids = []
            warehouse_ids = self.pool.get('stock.warehouse').search(cr, uid, [], context=context)
            for warehouse in self.pool.get('stock.warehouse').browse(cr, uid, warehouse_ids, context=context):
                location_ids.append(warehouse.lot_stock_id.id)

        #lots
        prodlot_id = context.get('prodlot_id', False)
        prodlot_condition_str = prodlot_id and ('and prodlot_id = ' + str(prodlot_id)) or ''

        if 'available' in what:
            if prodlot_condition_str!='':
                prodlot_ids = []
                prodlot_ids = self.pool.get('stock.production.lot').search(cr, uid,[('id','=',prodlot_id),('blocked','!=',False)],context=context)
                prodlot_condition_str = prodlot_ids and ('and prodlot_id = ' + str(prodlot_ids[0])) or ''
            else:
                prodlot_ids = []
                try: 
                    prodlot_ids = self.pool.get('stock.production.lot').search(cr, uid,[('product_id','in',ids),('blocked','=',True)],context=context)
                    if len(prodlot_ids)>1: prodlot_condition_str = 'and (prodlot_id not in '+ str(tuple(prodlot_ids))+' or prodlot_id IS NULL)'
                    else: prodlot_condition_str = 'and (prodlot_id <> '+ str(prodlot_ids[0])+' or prodlot_id IS NULL)'
                except: 
                    lot_exception = 1

        # build the list of ids of children of the location given by id
        if context.get('compute_child',True):
            child_location_ids = self.pool.get('stock.location').search(cr, uid, [('location_id', 'child_of', location_ids)])
            location_ids = child_location_ids or location_ids
        else:
            location_ids = location_ids

        uoms_o = {}
        product2uom = {}
        for product in self.browse(cr, uid, ids, context=context):
            product2uom[product.id] = product.uom_id.id
            uoms_o[product.uom_id.id] = product.uom_id

        results = []
        results2 = []

        from_date = context.get('from_date',False)
        to_date = context.get('to_date',False)
        date_str = False
        date_values = False
        where = [tuple(location_ids),tuple(location_ids),tuple(ids),tuple(states)]
        if from_date and to_date:
            date_str = "date>=%s and date<=%s"
            where.append(tuple([from_date]))
            where.append(tuple([to_date]))
        elif from_date:
            date_str = "date>=%s"
            date_values = [from_date]
        elif to_date:
            date_str = "date<=%s"
            date_values = [to_date]
        if date_values:
            where.append(tuple(date_values))
        date_condition_str = (date_str and 'and '+date_str+' ' or '')

    # TODO: write querys in more ORM way?

        if 'in' in what:
            # all moves from a location out of the set to a location in the set
            if 'available' in what:
                try: location_ids = self.pool.get('stock.location').search(cr, uid, [('id','in',where[1]),('blocked','=',False)], context=context)
                except:location_ids = ()
                try: product_ids = self.pool.get('product.product').search(cr, uid, [('id','in',where[2]),('blocked','=',False)], context=context)
                except:product_ids = ()
                if len(location_ids)>0: where[1]=tuple(location_ids)
                else: where[1]=(0,0)
                if len(product_ids)>0: where[2]=tuple(product_ids)
                else: where[2]=(0,0)

                query_string = 'select sum(product_qty), product_id, product_uom '\
                    'from stock_move '\
                    'where location_id NOT IN %s '\
                    'and location_dest_id IN %s '\
                    'and product_id IN %s '\
                    '' + prodlot_condition_str + ' and state IN %s '\
                    '' + date_condition_str +' '\
                    'group by product_id,product_uom'
            else:
                query_string = 'select sum(product_qty), product_id, product_uom '\
                    'from stock_move '\
                    'where location_id NOT IN %s '\
                    'and location_dest_id IN %s '\
                    'and product_id IN %s '\
                    '' + prodlot_condition_str + ' and state IN %s '\
                    '' + date_condition_str +' '\
                    'group by product_id,product_uom'

            cr.execute(query_string,tuple(where))
            results = cr.fetchall()
        if 'out' in what:
            # all moves from a location in the set to a location out of the set
            if 'available' in what:
                try: location_ids = self.pool.get('stock.location').search(cr, uid, [('id','in',where[0]),('blocked','=',False)], context=context)
                except:location_ids = ()
                try: product_ids = self.pool.get('product.product').search(cr, uid, [('id','in',where[2]),('blocked','=',False)], context=context)
                except:product_ids = ()
                if len(location_ids)>0: where[0]=tuple(location_ids)
                else: where[0]=(0,0)
                if len(product_ids)>0: where[2]=tuple(product_ids)
                else: where[2]=(0,0)
                query_string='select sum(product_qty), product_id, product_uom '\
                    'from stock_move '\
                    'where location_id IN %s '\
                    'and location_dest_id NOT IN %s '\
                    'and product_id  IN %s '\
                    '' + prodlot_condition_str + ' and state IN %s '\
                    '' + date_condition_str +' '\
                    'group by product_id,product_uom'
            else:
                query_string='select sum(product_qty), product_id, product_uom '\
                    'from stock_move '\
                    'where location_id IN %s '\
                    'and location_dest_id NOT IN %s '\
                    'and product_id  IN %s '\
                    '' + prodlot_condition_str + ' and state IN %s '\
                    '' + date_condition_str +' '\
                    'group by product_id,product_uom'
            cr.execute(query_string,tuple(where))
            results2 = cr.fetchall()

        uom_obj = self.pool.get('product.uom')
        uoms = map(lambda x: x[2], results) + map(lambda x: x[2], results2)
        if context.get('uom', False): uoms += [context['uom']]
        uoms = filter(lambda x: x not in uoms_o.keys(), uoms)
        if uoms:
            uoms = uom_obj.browse(cr, uid, list(set(uoms)), context=context)
            for o in uoms: uoms_o[o.id] = o
        #TOCHECK: before change uom of product, stock move line are in old uom.
        context.update({'raise-exception': False})       
        for amount, prod_id, prod_uom in results:
            amount = uom_obj._compute_qty_obj(cr, uid, uoms_o[prod_uom], amount,
                     uoms_o[context.get('uom', False) or product2uom[prod_id]], context=context)
            res[prod_id] += amount
        for amount, prod_id, prod_uom in results2:
            amount = uom_obj._compute_qty_obj(cr, uid, uoms_o[prod_uom], amount,
                    uoms_o[context.get('uom', False) or product2uom[prod_id]], context=context)
            res[prod_id] -= amount
        return res

    def _virtual_available_qty(self, cr, uid, ids, field_names=None, arg=False, context=None):
        if not field_names: field_names = []
        if context is None: context = {}
        res = dict.fromkeys(ids, '')
        c = context.copy()
        c.update({'product_id':ids[0]})
        #TODO: Multicompany? Various inlungs? Multiple warehouses?
        inlung_ids = self.pool.get('stock.location').search(cr, uid, [('type','=','inlung')], context=c)
        warehouse_ids = self.pool.get('stock.warehouse').search(cr, uid, [], context=c)        
        stock_id = self.pool.get('stock.warehouse').browse(cr, uid, warehouse_ids[0], context=c).lot_stock_id
        if inlung_ids != []: location_obj = self.pool.get('stock.location').browse(cr, uid, inlung_ids, context=c)[0]
        for obj in self.browse(cr, uid, ids, context=context):
            if obj.virtual_available_tmp > 0 and inlung_ids!=[] and stock_id.id == context.get('location'):
                res[obj.id] = obj.virtual_available_tmp - location_obj.stock_real
            else: res[obj.id] = obj.virtual_available_tmp
        return res

    def _product_available(self, cr, uid, ids, field_names=None, arg=False, context=None):
        """ Finds the incoming and outgoing quantity of product.
        @return: Dictionary of values
        """
        if not field_names:
            field_names = []
        if context is None:
            context = {}
        res = {}
        for id in ids:
            res[id] = {}.fromkeys(field_names, 0.0)
        for f in field_names:
            c = context.copy()
            if f == 'qty_available':
                c.update({ 'states': ('done',), 'what': ('in', 'out') })
            if f == 'virtual_available':
                c.update({ 'states': ('confirmed','waiting','assigned','done'), 'what': ('in', 'out') })
            if f == 'incoming_qty':
                c.update({ 'states': ('confirmed','waiting','assigned'), 'what': ('in',) })
            if f == 'outgoing_qty':
                c.update({ 'states': ('confirmed','waiting','assigned'), 'what': ('out',) })
            if f == 'real_available_qty':
                c.update({ 'states': ('done',), 'what': ('in', 'out', 'available') })
            if f == 'virtual_available_tmp':
                c.update({ 'states': ('confirmed','waiting','assigned','done'), 'what': ('in', 'out', 'available') })
            if f == 'virtual_available_qty':
                c.update({ 'states': ('confirmed','waiting','assigned','done'), 'what': ('in', 'out', 'available') })
            stock = self.get_product_available(cr, uid, ids, context=c)
            for id in ids:
                res[id][f] = stock.get(id, 0.0)
        return res

    _columns = {
        'real_available_qty': fields.function(_product_available, method=True, type='float', string='Real Quantity Available', help="Real stock without locked units.", multi=True, digits_compute=dp.get_precision('Product UoM'), ),
        'virtual_available_tmp': fields.function(_product_available, method=True,  type='float', string='Virtual Quantity Available', help="Virtual stock without locked units.", multi=True, digits_compute=dp.get_precision('Product UoM')),
        #'virtual_available_qty': fields.function(_product_available, method=True,  type='float', string='Virtual Quantity Available', help="Virtual stock without locked units.", multi=True, digits_compute=dp.get_precision('Product UoM')),
        'virtual_available_qty': fields.function(_virtual_available_qty, method=True,  type='float', string='Virtual Quantity Available', help="Virtual stock without locked units.", digits_compute=dp.get_precision('Product UoM')),
    }

available_product_stock()

# vim:expandtab:smartindent:tabstop=4:softtabstop=4:shiftwidth=4:
